/*

Velocity
Version 1.1
Written by Gabriel Anderson (Zettablade)
Copyright 2006, 2007 Gabriel Anderson

	This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

!!!DISCLAIMERS!!!
!!!DO NOT CLAIM VELOCITY AS YOUR OWN!!! 
!!!DO NOT USE VELOCITY TO MAKE MALICIOUS CONTENT!!!
!!!DO NOT USE VELOCITY FOR ANY COMMERCIAL PURPOSES WITHOUT EXPRESS PERMISSION FROM THE AUTHOR!!!
!!!DISCLAMERS!!!

*/

//---------------------------------------------------------------------------------//
// Includes                                                                        //
//---------------------------------------------------------------------------------//

#include "vlib.h"

//---------------------------------------------------------------------------------//
// Callbacks                                                                       //
//---------------------------------------------------------------------------------//

// Exit callback
static int exit_callback(int arg1, int arg2, void *common) {
	sceKernelExitGame();
	return 0;
}

// Callback thread 
static int CallbackThread(SceSize args, void *argp) {
	int cbid;

	cbid = sceKernelCreateCallback("Exit Callback", exit_callback, NULL);
	sceKernelRegisterExitCallback(cbid);

	sceKernelSleepThreadCB();

	return 0;
}

// Sets up the callback thread and returns its thread id
static int SetupCallbacks(void) {
	int thid = 0;

	thid = sceKernelCreateThread("update_thread", CallbackThread, 0x11, 0xFA0, 0, 0);
	if(thid >= 0)
		sceKernelStartThread(thid, 0, 0);

	return thid;
}

//---------------------------------------------------------------------------------//
// VRam stuff                                                                      //
//---------------------------------------------------------------------------------//

static unsigned int staticOffset = 0;

static unsigned int getMemorySize(unsigned int width, unsigned int height, unsigned int psm)
{
	switch (psm)
	{
		case GU_PSM_T4:
			return (width * height) >> 1;

		case GU_PSM_T8:
			return width * height;

		case GU_PSM_5650:
		case GU_PSM_5551:
		case GU_PSM_4444:
		case GU_PSM_T16:
			return 2 * width * height;

		case GU_PSM_8888:
		case GU_PSM_T32:
			return 4 * width * height;

		default:
			return 0;
	}
}

void* getStaticVramBuffer(unsigned int width, unsigned int height, unsigned int psm)
{
	unsigned int memSize = getMemorySize(width,height,psm);
	void* result = (void*)staticOffset;
	staticOffset += memSize;

	return result;
}

//---------------------------------------------------------------------------------//
// General                                                                         //
//---------------------------------------------------------------------------------//

void* fbp0;
void* fbp1;
void* zbp;
unsigned int __attribute__((aligned(16))) DList[262144];
int fps = 0, lastFps = 0;		// for calculating the frames per second

u32 tickResolution;
u64 fpsTickNow;
u64 fpsTickLast;

CONTROLLER buttonInit;
SceCtrlData ctrl;

const int BUF_WIDTH = 512;
const int SCR_WIDTH = 480;
const int SCR_HEIGHT = 272;
const int FRAMEBUFFER_SIZE = 512*272*4;

const int PRINT_BUFF = 512;

const int IN_RAM = 0;
const int IN_VRAM = 1;

const int OFF = 0;
const int ON = 1;

const int DRAW_2D = 0;
const int DRAW_3D = 1;

const int PI = 3.14159;

short drawingStarted = 0;

void initVLib(u32 clearColor)
{
	pspDebugScreenInit();
	SetupCallbacks();
	sceKernelDcacheWritebackAll();

	sceCtrlSetSamplingCycle(0);
	sceCtrlSetSamplingMode(1);

	button = &buttonInit;

	sceRtcGetCurrentTick( &fpsTickLast );
	tickResolution = sceRtcGetTickResolution();
	
	fbp0 = (void*) getStaticVramBuffer(BUF_WIDTH,SCR_HEIGHT,GU_PSM_8888);
	fbp1 = (void*) getStaticVramBuffer(BUF_WIDTH,SCR_HEIGHT,GU_PSM_8888);
	zbp = (void*) getStaticVramBuffer(BUF_WIDTH,SCR_HEIGHT,GU_PSM_4444);
	
	// Setup GU

	sceGuInit();
	sceGuStart(GU_DIRECT,DList);

	sceGuDrawBuffer(GU_PSM_8888,fbp0,BUF_WIDTH);
	sceGuDispBuffer(SCR_WIDTH,SCR_HEIGHT,fbp1,BUF_WIDTH);
	sceGuDepthBuffer(zbp,BUF_WIDTH);

	sceKernelDcacheWritebackAll();

	sceGuOffset(2048 - (SCR_WIDTH/2),2048 - (SCR_HEIGHT/2));
	sceGuViewport(2048,2048,SCR_WIDTH,SCR_HEIGHT);
	sceGuDepthRange(65535,0);

	// Setup the scissor test
	sceGuScissor(0,0,SCR_WIDTH,SCR_HEIGHT);
	sceGuEnable(GU_SCISSOR_TEST);
	// Setup the depth testing
	sceGuDepthFunc(GU_GEQUAL);
	sceGuEnable(GU_DEPTH_TEST);
	// Enable clipping
	sceGuEnable(GU_CLIP_PLANES);
	// Setup face culling
	//sceGuFrontFace(GU_CW);
	//sceGuEnable(GU_CULL_FACE);
	// Setup shading (not lighting!)
	sceGuShadeModel( GU_SMOOTH );
	// Setup textures
	sceGuTexMode(GU_PSM_8888,0,0,1);
	sceGuTexFunc(GU_TFX_REPLACE, GU_TCC_RGBA);
	// Enable 2D textures
	sceGuEnable(GU_TEXTURE_2D);
	// Do alpha blending
	sceGuBlendFunc(GU_ADD,GU_SRC_ALPHA,GU_ONE_MINUS_SRC_ALPHA,0,0);
	sceGuEnable(GU_BLEND);
	// Setup the alpha test
	sceGuAlphaFunc(GU_GREATER, 0, 0xff);
	sceGuEnable(GU_ALPHA_TEST);
	// Setup the color/depth for clearing the screen
	sceGuClearColor(clearColor);
	sceGuClearDepth(0);

	sceGuFinish();
	sceGuSync(0,0);

	sceDisplayWaitVblankStart();
	sceGuDisplay(GU_TRUE);

	sceKernelDcacheWritebackAll();
}

void uninitVLib()
{
	free(fbp0);
	free(fbp1);
	free(zbp);
	free(DList);
	sceGuTerm();
}

void startDrawing()
{
	if(drawingStarted)
		return;
	sceGuStart(GU_DIRECT, DList);
	drawingStarted = 1;
}

void endDrawing()
{
	if(!drawingStarted)
		return;
	sceGuFinish();
	sceGuSync(0,0);
	drawingStarted = 0;
}

void syncFrame()
{
	if(drawingStarted)
		return;
	asm __volatile__ ("\nsyscall 0x2147\n");
	fbp0 = sceGuSwapBuffers();
}

void flipScreen()
{
	if(drawingStarted)
		return;
	fbp0 = sceGuSwapBuffers();
}

void clearScreen()
{
	if(!drawingStarted)
		return;
	sceGuClear(GU_COLOR_BUFFER_BIT|GU_DEPTH_BUFFER_BIT);
}

void waitVSync()
{
	if(drawingStarted)
		return;
	asm __volatile__ ("\nsyscall 0x2147\n");
}

void setDrawMode(short drawMode)
{
	if(drawMode) {
		sceGumMatrixMode(GU_PROJECTION);
		sceGumLoadIdentity();
		sceGumPerspective( 75.0, 16.0/9.0, .5, 1000.0);
		
		sceGumMatrixMode(GU_VIEW);
		sceGumLoadIdentity();
		
		sceGumMatrixMode(GU_MODEL);
		sceGumLoadIdentity();
	} else {
		sceGumMatrixMode(GU_PROJECTION);
		sceGumLoadIdentity();
		sceGumOrtho(0, 480, 272, 0, -1, 1);
		
		sceGumMatrixMode(GU_VIEW);
		sceGumLoadIdentity();
 		
		sceGumMatrixMode(GU_MODEL);
		sceGumLoadIdentity();
	}
}

//---------------------------------------------------------------------------------//
// Image                                                                           //
//---------------------------------------------------------------------------------//

static int getPow2(int value)
{
	int b = value;
	int n;
	for (n = 0; b != 0; n++)
		b >>= 1;

	b = 1 << n;

	if (b == 2 * value)
		b >>= 1;
	return b;
}

Image *loadImage(const char *filename, int location)
{
	Image *img = (Image*) malloc(sizeof(Image));
	const char *suffix = strrchr(filename, '.');
	FILE *fp;
	
	// open the image file
	if ((fp = fopen(filename, "rb")) == NULL)
		return NULL;
		
	if(strcmp(suffix,".png") == 0) {
		png_structp png_ptr;
		png_infop info_ptr;
		unsigned int sig_read = 0;
		png_uint_32 width, height;
		int bit_depth, color_type, interlace_type, x, y;
		
		png_ptr = png_create_read_struct(PNG_LIBPNG_VER_STRING, NULL, NULL, NULL);
		info_ptr = png_create_info_struct(png_ptr);
		png_init_io(png_ptr, fp);
		png_set_sig_bytes(png_ptr, sig_read);
		png_read_info(png_ptr, info_ptr);
		png_get_IHDR(png_ptr, info_ptr, &width, &height, &bit_depth, &color_type, &interlace_type, int_p_NULL, int_p_NULL);
	
		img->pixelWidth = width;
		img->pixelHeight = height;
		img->realWidth = getPow2(width);
		img->realHeight = getPow2(height);
	
		img->totalSize = img->realWidth * img->realHeight * sizeof(u32);
	
		png_set_strip_16(png_ptr);
		png_set_packing(png_ptr);
		if (color_type == PNG_COLOR_TYPE_PALETTE)
			png_set_palette_to_rgb(png_ptr);
		if (color_type == PNG_COLOR_TYPE_GRAY && bit_depth < 8)
			png_set_gray_1_2_4_to_8(png_ptr);
		if (png_get_valid(png_ptr, info_ptr, PNG_INFO_tRNS))
			png_set_tRNS_to_alpha(png_ptr);
		png_set_filler(png_ptr, 0xff, PNG_FILLER_AFTER);
	
		img->data = (u32*)malloc(img->totalSize);
	
		if(location) {
			sceGuCopyImage(GU_PSM_8888,0,0,img->realWidth,img->realHeight,img->realWidth,img->data,0,0,BUF_WIDTH,(void*)(((unsigned int)fbp0)+0x4000000));
			sceGuTexSync();
		}
	
		u32* line = (u32*) malloc(width * 4);
		for(y = 0; y < (int)height; y++) {
			png_read_row(png_ptr, (u8*) line, png_bytep_NULL);
			for(x = 0; x < (int)width; x++) {
				u32 color = line[x];
				img->data[x + y * (int)img->realWidth] =  color;
			}
		}
		free(line);
		
		png_read_end(png_ptr, info_ptr);
		png_destroy_read_struct(&png_ptr, &info_ptr, png_infopp_NULL);
	} else if(strcmp(suffix,".bmp") == 0) {
		// bitmap file loader goes here
	} else if(strcmp(suffix,".tga") == 0) {
		unsigned char ucharBad;	// garbage data
		short int	sintBad;		// garbage data
		int colorMode;			// 4 for RGBA, 3 for RGB
		long imageIdx;			// counter variable
		unsigned char colorSwap;	// swap variable
		unsigned char imageTypeCode;
		unsigned char bitCount;
		
		// read first two bytes of garbage
		fread(&ucharBad, sizeof(unsigned char), 1, fp);
		fread(&ucharBad, sizeof(unsigned char), 1, fp);
	 
		// read in the image type
		fread(&imageTypeCode, sizeof(unsigned char), 1, fp);
	 
		// for our purposes, the image type should be either a 2 or a 3
		if ((imageTypeCode != 2) && (imageTypeCode != 3))
		{
			fclose(fp);
			return NULL;
		}
	 
		// read 13 bytes of garbage data
		fread(&sintBad, sizeof(short int), 1, fp);
		fread(&sintBad, sizeof(short int), 1, fp);
		fread(&ucharBad, sizeof(unsigned char), 1, fp);
		fread(&sintBad, sizeof(short int), 1, fp);
		fread(&sintBad, sizeof(short int), 1, fp);
	 
		// read image dimensions
		fread(&img->pixelWidth, sizeof(short int), 1, fp);
		fread(&img->pixelHeight, sizeof(short int), 1, fp);
		
		img->realWidth = getPow2(img->pixelWidth);
		img->realHeight = getPow2(img->pixelHeight);
	 
		// read bit depth
		fread(&bitCount, sizeof(unsigned char), 1, fp);
	 
		// read garbage
		fread(&ucharBad, sizeof(unsigned char), 1, fp);
	 
		// colormode -> 3 = BGR, 4 = BGRA
		colorMode = bitCount / 8;
		img->totalSize = img->realWidth * img->realHeight * colorMode;
	 
		// allocate memory for image data
		img->data = (u32*)malloc(sizeof(u32)*img->totalSize);
	 
		// read image data
		fread(img->data, sizeof(u32), img->totalSize, fp);
		
		// change BGR to RGB
		for (imageIdx = 0; imageIdx < img->totalSize; imageIdx += colorMode)
		{
			colorSwap = img->data[imageIdx];
			img->data[imageIdx] = img->data[imageIdx+2];
			img->data[imageIdx + 2] = colorSwap;
		}
	} else if((strcmp(suffix,".jpeg") == 0) || (strcmp(suffix,".jpg") == 0) || (strcmp(suffix,".jpe") == 0)) {
	}
	
	// close the file
	fclose(fp);
	
	// Swizzle the texture
	int rowblocks = ((img->realWidth * 4)/16);
	int rowblocks_add = (rowblocks-1) * 128;
	unsigned int block_address = 0;
	unsigned int *src = (unsigned int*)img->data;
	u8* temp = (u8*)malloc(img->totalSize);

	int j;
	for(j=0; j<img->realHeight; j++, block_address+=16) {
		unsigned int *block = (unsigned int*)((unsigned int)&temp[block_address]|0x40000000);
		int i;
		for(i=0; i < rowblocks; i++) {
			*block++ = *src++;
			*block++ = *src++;
			*block++ = *src++;
			*block++ = *src++;
			block+=28;
		}
		if((j&0x7)==0x7)
			block_address += rowblocks_add;
	}

	free(img->data);
	img->data = (u32*)temp;
	
	// Clear bad data from the cache
	sceKernelDcacheWritebackAll();
	
	return img;
}

void unloadImage(Image *img)
{
	free(img->data);
	free(img);
}

void drawImage(float x, float y, float width, float height, Image *img)
{	
	// Set some variables
	if(width==0 || height==0)
		return;
	float slice = img->pixelWidth/64;
	float x_width = width/slice;
	float u_width = img->pixelWidth/slice;
	float cur_u = 0;
	float cur_x = 0;
	
	// Move to the desired point
	ScePspFVector3 move = {x, y, 0};
	sceGumTranslate(&move);
	
	// Set the texture so the u/v is right
	float u = 1.0f / ((float)img->pixelWidth);
	float v = 1.0f / ((float)img->pixelHeight);
	sceGuTexScale(u, v);
	
	// Set texture
	sceGuTexImage(0,img->realWidth,img->realHeight,img->realWidth,(void*)img->data);
	
	// Blit stripping routine
	float ii;
	for(ii = slice; ii--;) {
		TP_Vertex_3D* vertices = (TP_Vertex_3D*)sceGuGetMemory(2 * sizeof(TP_Vertex_3D));
		
		vertices[0].u = cur_u;
		vertices[0].v = 0;
		vertices[0].x = cur_x;
		vertices[0].y = 0;
		vertices[0].z = 0;
		
		cur_u += u_width;
		cur_x += x_width;
		
		vertices[1].u = cur_u;
		vertices[1].v = img->pixelHeight;
		vertices[1].x = cur_x;
		vertices[1].y = height;
		vertices[1].z = 0;
		
		sceGumDrawArray(GU_SPRITES, GU_TEXTURE_32BITF|GU_VERTEX_32BITF|GU_TRANSFORM_3D, 2, 0, vertices);
	}
	// Reset the texture scale
	sceGuTexScale(1.0, 1.0);
}

void drawImageCenter(float x, float y, float width, float height, float centerX, float centerY, Image *img)
{	
	if(width==0 || height==0)
		return;
	float slice = img->pixelWidth/64;
	float x_width = width/slice;
	float u_width = img->pixelWidth/slice;
	float cur_u = 0;
	float cur_x = 0;
	
	sceGumMatrixMode(GU_VIEW);
	sceGumLoadIdentity();
	
	ScePspFVector3 center = {centerX, centerY, 0};
	sceGumTranslate(&center);
	
	sceGumMatrixMode(GU_MODEL);
	sceGumLoadIdentity();
	
	ScePspFVector3 move = {x, y, 0};
	sceGumTranslate(&move);
	
	float u = 1.0f / ((float)img->pixelWidth);
	float v = 1.0f / ((float)img->pixelHeight);
	sceGuTexScale(u, v);
	
	sceGuTexImage(0,img->realWidth,img->realHeight,img->realWidth,(void*)img->data);
	
	float ii;
	for(ii = slice; ii--;) {
		TP_Vertex_3D* vertices = (TP_Vertex_3D*)sceGuGetMemory(2 * sizeof(TP_Vertex_3D));
		
		vertices[0].u = cur_u;
		vertices[0].v = 0;
		vertices[0].x = cur_x;
		vertices[0].y = 0;
		vertices[0].z = 0;
		
		cur_u += u_width;
		cur_x += x_width;
		
		vertices[1].u = cur_u;
		vertices[1].v = img->pixelHeight;
		vertices[1].x = cur_x;
		vertices[1].y = height;
		vertices[1].z = 0;
		
		sceGumDrawArray(GU_SPRITES, GU_TEXTURE_32BITF|GU_VERTEX_32BITF|GU_TRANSFORM_3D, 2, 0, vertices);
	}
	
	// Reset the view matrix, model matrix, and the texture scale
	sceGumMatrixMode(GU_VIEW);
	sceGumLoadIdentity();
	
	sceGumMatrixMode(GU_MODEL);
	sceGumLoadIdentity();
	
	sceGuTexScale(1.0, 1.0);
}

void drawImageAngle(float x, float y, float width, float height, float angle, Image *img)
{	
	if(width==0 || height==0)
		return;
	float slice = img->pixelWidth/64;
	float x_width = width/slice;
	float u_width = img->pixelWidth/slice;
	float cur_u = 0;
	float cur_x = 0;
	
	sceGumMatrixMode(GU_VIEW);
	sceGumLoadIdentity();
	
	ScePspFVector3 center = {width/2, height/2, 0};
	sceGumTranslate(&center);
	
	sceGumRotateZ(angle * (GU_PI/180.0f));
	
	sceGumMatrixMode(GU_MODEL);
	sceGumLoadIdentity();
	
	ScePspFVector3 move = {x, y, 0};
	sceGumTranslate(&move);
	
	float u = 1.0f / ((float)img->pixelWidth);
	float v = 1.0f / ((float)img->pixelHeight);
	sceGuTexScale(u, v);
	
	sceGuTexImage(0,img->realWidth,img->realHeight,img->realWidth,(void*)img->data);
	
	float ii;
	for(ii = slice; ii--;) {
		TP_Vertex_3D* vertices = (TP_Vertex_3D*)sceGuGetMemory(4 * sizeof(TP_Vertex_3D));
		
		vertices[0].u = cur_u;
		vertices[0].v = 0;
		vertices[0].x = cur_x;
		vertices[0].y = 0;
		vertices[0].z = 0;
		
		vertices[2].u = cur_u;
		vertices[2].v = img->pixelHeight;
		vertices[2].x = cur_x;
		vertices[2].y = height;
		vertices[2].z = 0;
		
		cur_u += u_width;
		cur_x += x_width;
		
		vertices[1].u = cur_u;
		vertices[1].v = 0;
		vertices[1].x = cur_x;
		vertices[1].y = 0;
		vertices[1].z = 0;
		
		vertices[3].u = cur_u;
		vertices[3].v = img->pixelHeight;
		vertices[3].x = cur_x;
		vertices[3].y = height;
		vertices[3].z = 0;
		
		sceGumDrawArray(GU_TRIANGLE_STRIP, GU_TEXTURE_32BITF|GU_VERTEX_32BITF|GU_TRANSFORM_3D, 4, 0, vertices);
	}
	
	// Reset the view matrix, model matrix, and the texture scale
	sceGumMatrixMode(GU_VIEW);
	sceGumLoadIdentity();
	
	sceGumMatrixMode(GU_MODEL);
	sceGumLoadIdentity();
	
	sceGuTexScale(1.0, 1.0);
}

void drawImageAngleCenter(float x, float y, float width, float height, float centerX, float centerY, float angle, Image *img)
{	
	if(width==0 || height==0)
		return;
	float slice = img->pixelWidth/64;
	float x_width = width/slice;
	float u_width = img->pixelWidth/slice;
	float cur_u = 0;
	float cur_x = 0;
	
	sceGumMatrixMode(GU_VIEW);
	sceGumLoadIdentity();
	
	ScePspFVector3 center = {centerX, centerY, 0};
	sceGumTranslate(&center);
	
	sceGumRotateZ(angle * (GU_PI/180.0f));
	
	sceGumMatrixMode(GU_MODEL);
	sceGumLoadIdentity();
	
	ScePspFVector3 move = {x, y, 0};
	sceGumTranslate(&move);
	
	float u = 1.0f / ((float)img->pixelWidth);
	float v = 1.0f / ((float)img->pixelHeight);
	sceGuTexScale(u, v);
	
	sceGuTexImage(0,img->realWidth,img->realHeight,img->realWidth,(void*)img->data);
	
	float ii;
	for(ii = slice; ii--;) {
		TP_Vertex_3D* vertices = (TP_Vertex_3D*)sceGuGetMemory(4 * sizeof(TP_Vertex_3D));
		
		vertices[0].u = cur_u;
		vertices[0].v = 0;
		vertices[0].x = cur_x;
		vertices[0].y = 0;
		vertices[0].z = 0;
		
		vertices[2].u = cur_u;
		vertices[2].v = img->pixelHeight;
		vertices[2].x = cur_x;
		vertices[2].y = height;
		vertices[2].z = 0;
		
		cur_u += u_width;
		cur_x += x_width;
		
		vertices[1].u = cur_u;
		vertices[1].v = 0;
		vertices[1].x = cur_x;
		vertices[1].y = 0;
		vertices[1].z = 0;
		
		vertices[3].u = cur_u;
		vertices[3].v = img->pixelHeight;
		vertices[3].x = cur_x;
		vertices[3].y = height;
		vertices[3].z = 0;
		
		sceGumDrawArray(GU_TRIANGLE_STRIP, GU_TEXTURE_32BITF|GU_VERTEX_32BITF|GU_TRANSFORM_3D, 4, 0, vertices);
	}
	
	// Reset the view matrix, model matrix, and the texture scale
	sceGumMatrixMode(GU_VIEW);
	sceGumLoadIdentity();
	
	sceGumMatrixMode(GU_MODEL);
	sceGumLoadIdentity();
	
	sceGuTexScale(1.0, 1.0);
}

void drawImageCrop(float x, float y, float width, float height, float beginX, float beginY, float endX, float endY, Image *img)
{
	if(width==0 || height==0)
		return;
	float slice = endX/64;
	float x_width = width/slice;
	float u_width = endX/slice;
	float cur_u = beginX;
	float cur_x = 0;
	
	ScePspFVector3 move = {x, y, 0};
	sceGumTranslate(&move);
	
	float u = 1.0f / ((float)img->pixelWidth);
	float v = 1.0f / ((float)img->pixelHeight);
	sceGuTexScale(u, v);
	
	sceGuTexImage(0,img->realWidth,img->realHeight,img->realWidth,(void*)img->data);
	
	float ii;
	for(ii = slice; ii--;) {
		TP_Vertex_3D* vertices = (TP_Vertex_3D*)sceGuGetMemory(2 * sizeof(TP_Vertex_3D));
		
		vertices[0].u = cur_u;
		vertices[0].v = beginY;
		vertices[0].x = cur_x;
		vertices[0].y = 0;
		vertices[0].z = 0;
		
		cur_u += u_width;
		cur_x += x_width;
		
		vertices[1].u = cur_u;
		vertices[1].v = endY;
		vertices[1].x = cur_x;
		vertices[1].y = height;
		vertices[1].z = 0;
		
		sceGumDrawArray(GU_SPRITES, GU_TEXTURE_32BITF|GU_VERTEX_32BITF|GU_TRANSFORM_3D, 2, 0, vertices);
	}
	
	// Reset the texture scale
	sceGuTexScale(1.0, 1.0);
}

void drawImageCropCenter(float x, float y, float width, float height, float centerX, float centerY, float beginX, float beginY, float endX, float endY, Image *img)
{
	if(width==0 || height==0)
		return;
	float slice = endX/64;
	float x_width = width/slice;
	float u_width = endX/slice;
	float cur_u = beginX;
	float cur_x = 0;
	
	sceGumMatrixMode(GU_VIEW);
	sceGumLoadIdentity();
	
	ScePspFVector3 center = {centerX, centerY, 0};
	sceGumTranslate(&center);
	
	sceGumMatrixMode(GU_MODEL);
	sceGumLoadIdentity();
	
	ScePspFVector3 move = {x, y, 0};
	sceGumTranslate(&move);
	
	float u = 1.0f / ((float)img->pixelWidth);
	float v = 1.0f / ((float)img->pixelHeight);
	sceGuTexScale(u, v);
	
	sceGuTexImage(0,img->realWidth,img->realHeight,img->realWidth,(void*)img->data);
	
	float ii;
	for(ii = slice; ii--;) {
		TP_Vertex_3D* vertices = (TP_Vertex_3D*)sceGuGetMemory(2 * sizeof(TP_Vertex_3D));
		
		vertices[0].u = cur_u;
		vertices[0].v = beginY;
		vertices[0].x = cur_x;
		vertices[0].y = 0;
		vertices[0].z = 0;
		
		cur_u += u_width;
		cur_x += x_width;
		
		vertices[1].u = cur_u;
		vertices[1].v = endY;
		vertices[1].x = cur_x;
		vertices[1].y = height;
		vertices[1].z = 0;
		
		sceGumDrawArray(GU_SPRITES, GU_TEXTURE_32BITF|GU_VERTEX_32BITF|GU_TRANSFORM_3D, 2, 0, vertices);
	}
	
	// Reset the view matrix, model matrix, and the texture scale
	sceGumMatrixMode(GU_VIEW);
	sceGumLoadIdentity();
	
	sceGumMatrixMode(GU_MODEL);
	sceGumLoadIdentity();
	
	sceGuTexScale(1.0, 1.0);
}

void drawImageCropAngle(float x, float y, float width, float height, float beginX, float beginY, float endX, float endY, float angle, Image *img)
{
	if(width==0 || height==0)
		return;
	float slice = endX/64;
	float x_width = width/slice;
	float u_width = endX/slice;
	float cur_u = beginX;
	float cur_x = 0;
	
	sceGumMatrixMode(GU_VIEW);
	sceGumLoadIdentity();
	
	ScePspFVector3 center = {width/2, height/2, 0};
	sceGumTranslate(&center);
	
	sceGumRotateZ(angle * (GU_PI/180.0f));
	
	sceGumMatrixMode(GU_MODEL);
	sceGumLoadIdentity();
	
	ScePspFVector3 move = {x, y, 0};
	sceGumTranslate(&move);
	
	float u = 1.0f / ((float)img->pixelWidth);
	float v = 1.0f / ((float)img->pixelHeight);
	sceGuTexScale(u, v);
	
	sceGuTexImage(0,img->realWidth,img->realHeight,img->realWidth,(void*)img->data);
	
	float ii;
	for(ii = slice; ii--;) {
		TP_Vertex_3D* vertices = (TP_Vertex_3D*)sceGuGetMemory(4 * sizeof(TP_Vertex_3D));
		
		vertices[0].u = cur_u;
		vertices[0].v = beginY;
		vertices[0].x = cur_x;
		vertices[0].y = 0;
		vertices[0].z = 0;
		
		vertices[2].u = cur_u;
		vertices[2].v = endY;
		vertices[2].x = cur_x;
		vertices[2].y = height;
		vertices[2].z = 0;
		
		cur_u += u_width;
		cur_x += x_width;
		
		vertices[1].u = cur_u;
		vertices[1].v = beginY;
		vertices[1].x = cur_x;
		vertices[1].y = 0;
		vertices[1].z = 0;
		
		vertices[3].u = cur_u;
		vertices[3].v = endY;
		vertices[3].x = cur_x;
		vertices[3].y = height;
		vertices[3].z = 0;
		
		sceGumDrawArray(GU_TRIANGLE_STRIP, GU_TEXTURE_32BITF|GU_VERTEX_32BITF|GU_TRANSFORM_3D, 4, 0, vertices);
	}
	
	// Reset the view matrix, model matrix, and the texture scale
	sceGumMatrixMode(GU_VIEW);
	sceGumLoadIdentity();
	
	sceGumMatrixMode(GU_MODEL);
	sceGumLoadIdentity();
	
	sceGuTexScale(1.0, 1.0);
}

void drawImageCropAngleCenter(float x, float y, float width, float height, float centerX, float centerY, float beginX, float beginY, float endX, float endY, float angle, Image *img)
{
	if(width==0 || height==0)
		return;
	float slice = endX/64;
	float x_width = width/slice;
	float u_width = endX/slice;
	float cur_u = beginX;
	float cur_x = 0;
	
	sceGumMatrixMode(GU_VIEW);
	sceGumLoadIdentity();
	
	ScePspFVector3 center = {centerX, centerY, 0};
	sceGumTranslate(&center);
	
	sceGumRotateZ(angle * (GU_PI/180.0f));
	
	sceGumMatrixMode(GU_MODEL);
	sceGumLoadIdentity();
	
	ScePspFVector3 move = {x, y, 0};
	sceGumTranslate(&move);
	
	float u = 1.0f / ((float)img->pixelWidth);
	float v = 1.0f / ((float)img->pixelHeight);
	sceGuTexScale(u, v);
	
	sceGuTexImage(0,img->realWidth,img->realHeight,img->realWidth,(void*)img->data);
	
	float ii;
	for(ii = slice; ii--;) {
		TP_Vertex_3D* vertices = (TP_Vertex_3D*)sceGuGetMemory(4 * sizeof(TP_Vertex_3D));
		
		vertices[0].u = cur_u;
		vertices[0].v = beginY;
		vertices[0].x = cur_x;
		vertices[0].y = 0;
		vertices[0].z = 0;
		
		vertices[2].u = cur_u;
		vertices[2].v = endY;
		vertices[2].x = cur_x;
		vertices[2].y = height;
		vertices[2].z = 0;
		
		cur_u += u_width;
		cur_x += x_width;
		
		vertices[1].u = cur_u;
		vertices[1].v = beginY;
		vertices[1].x = cur_x;
		vertices[1].y = 0;
		vertices[1].z = 0;
		
		vertices[3].u = cur_u;
		vertices[3].v = endY;
		vertices[3].x = cur_x;
		vertices[3].y = height;
		vertices[3].z = 0;
		
		sceGumDrawArray(GU_TRIANGLE_STRIP, GU_TEXTURE_32BITF|GU_VERTEX_32BITF|GU_TRANSFORM_3D, 4, 0, vertices);
	}
	
	// Reset the view matrix, model matrix, and the texture scale
	sceGumMatrixMode(GU_VIEW);
	sceGumLoadIdentity();
	
	sceGumMatrixMode(GU_MODEL);
	sceGumLoadIdentity();
	
	sceGuTexScale(1.0, 1.0);
}

/*void drawImage(float x, float y, float width, float height, Image *img)
{
	if(width==0 || height==0)
		return;
	float slice = img->pixelWidth/32;
	float x_width = width/slice;
	float u_width = img->pixelWidth/slice;
	float u0 = 0;

	sceGuTexImage(0,img->realWidth,img->realHeight,img->realWidth,(void*)img->data);

	float ii;
	for(ii = slice; ii--;) {
		TP_Vertex_3D* vertices = (TP_Vertex_3D*)sceGuGetMemory(2 * sizeof(TP_Vertex_3D));

		vertices[0].u = u0;
		vertices[0].v = 0;
		vertices[0].x = x;
		vertices[0].y = y;

		u0 += u_width;
		x += x_width;

		vertices[1].u = u0;
		vertices[1].v = img->pixelHeight;
		vertices[1].x = x;
		vertices[1].y = y + height;

		sceGuDisable(GU_DEPTH_TEST);
		sceGumDrawArray(GU_SPRITES, GU_TEXTURE_32BITF|GU_VERTEX_32BITF|GU_TRANSFORM_3D, 2, 0, vertices);
		sceGuEnable(GU_DEPTH_TEST);
	}
}

void drawImageAngle(int x, int y, int width, int height, float angle, Image *img)
{
	if(width==0 || height==0)
		return;
	float slice = img->pixelWidth/32;
	float x_width = width/slice;
	float u_width = img->pixelWidth/slice;
	int u0 = 0;

	sceGuTexImage(0,img->realWidth,img->realHeight,img->realWidth,(void*)img->data);

	float ii;
	for(ii = slice; ii--;) {
		TP_Vertex_2D* vertices = (TP_Vertex_2D*)sceGuGetMemory(2 * sizeof(TP_Vertex_2D));

		vertices[0].u = u0;
		vertices[0].v = 0;
		vertices[0].x = x;
		vertices[0].y = y;

		u0 += (short)u_width;
		x += (short)x_width;

		vertices[1].u = u0;
		vertices[1].v = img->pixelHeight;
		vertices[1].x = x;
		vertices[1].y = y + height;

		sceGuDisable(GU_DEPTH_TEST);
		sceGuDrawArray(GU_SPRITES, GU_TEXTURE_16BIT|GU_VERTEX_16BIT|GU_TRANSFORM_2D, 2, 0, vertices);
		sceGuEnable(GU_DEPTH_TEST);
	}
}

void drawImageUV(int x, int y, int width, int height, int u0, int u1, int v0, int v1, Image *img)
{
	if(width==0 || height==0)
		return;
	float slice = u1/32;
	float x_width = width/slice;
	float u_width = u1/slice;

	sceGuTexImage(0,img->realWidth,img->realHeight,img->realWidth,(void*)img->data);

	float ii;
	for(ii = slice; ii--;) {
		TP_Vertex_2D* vertices = (TP_Vertex_2D*)sceGuGetMemory(2 * sizeof(TP_Vertex_2D));

		vertices[0].u = u0;
		vertices[0].v = v0;
		vertices[0].x = x;
		vertices[0].y = y;

		u0 += (short)u_width;
		x += (short)x_width;

		vertices[1].u = u0;
		vertices[1].v = v1;
		vertices[1].x = x;
		vertices[1].y = y + height;

		sceGuDisable(GU_DEPTH_TEST);
		sceGuDrawArray(GU_SPRITES, GU_TEXTURE_16BIT|GU_VERTEX_16BIT|GU_TRANSFORM_2D, 2, 0, vertices);
		sceGuEnable(GU_DEPTH_TEST);
	}
}

void drawImageUVAngle(int x, int y, int width, int height, int u0, int u1, int v0, int v1, float angle, Image *img)
{
	if(width==0 || height==0)
		return;
	float slice = u1/32;
	float x_width = width/slice;
	float u_width = u1/slice;

	sceGuTexImage(0,img->realWidth,img->realHeight,img->realWidth,(void*)img->data);

	float ii;
	for(ii = slice; ii--;) {
		TP_Vertex_2D* vertices = (TP_Vertex_2D*)sceGuGetMemory(2 * sizeof(TP_Vertex_2D));

		vertices[0].u = u0;
		vertices[0].v = v0;
		vertices[0].x = x;
		vertices[0].y = y;

		u0 += (short)u_width;
		x += (short)x_width;

		vertices[1].u = u0;
		vertices[1].v = v1;
		vertices[1].x = x;
		vertices[1].y = y + height;

		sceGuDisable(GU_DEPTH_TEST);
		sceGuDrawArray(GU_SPRITES, GU_TEXTURE_16BIT|GU_VERTEX_16BIT|GU_TRANSFORM_2D, 2, 0, vertices);
		sceGuEnable(GU_DEPTH_TEST);
	}
}*/

void linearFilter(short state)
{
	sceGuStart(GU_DIRECT, DList);
	sceGuTexFilter(state, state);
	sceGuFinish();
	sceGuSync(0,0);
}

//---------------------------------------------------------------------------------//
// 3D Model                                                                        //
//---------------------------------------------------------------------------------//

static void ReadStr( FILE *f, char *string )
{
	do {
		fgets(string, 255, f);
	} while((string[0] == '/') || (string[0] == '\n'));
}

Model *loadModel(const char *filename)
{
	Model *mdl = (Model*) malloc(sizeof(Model));
	const char *suffix = strrchr(filename, '.');
	FILE *fp;
	if ((fp = fopen(filename, "rb")) == NULL)
			return NULL;

	if(strcmp(suffix,".3ds") == 0) {
	} else if(strcmp(suffix,".obj") == 0) {
	} else if(strcmp(suffix,".md2") == 0) {
	} else if(strcmp(suffix,".blend") == 0) {
	} else if(strcmp(suffix,".maya") == 0) {
	} else { // This is our .mdl file. It's only temp until I can get some of the larger, more complex files finished
		float x, y, z, u, v;
		int numOfVertices = 0;
		char oneLine[255];
		// Read one line
		ReadStr(fp, oneLine);
		// Find the number of triangles
		sscanf(oneLine, "verts %i\n", &mdl->vertTotal);
		
		// Allocate memory for the vertices
		mdl->verts = (TP_Vertex_3D*) malloc(mdl->vertTotal * sizeof(TP_Vertex_3D));
		//mdl->displayList = (unsigned int*) malloc(mdl->vertTotal);
		
		// Load the geometry line by line
		int loop;
		for(loop = 0; loop < mdl->vertTotal/3; loop++) {
			int vert;
			for(vert = 0; vert < 3; vert++) {
				ReadStr(fp, oneLine);
				sscanf( oneLine, "%f %f %f %f %f", &x, &y, &z, &u, &v );
				mdl->verts[numOfVertices].x = x;
				mdl->verts[numOfVertices].y = y;
				mdl->verts[numOfVertices].z = z;
				mdl->verts[numOfVertices].u = u;
				mdl->verts[numOfVertices].v = v;
				numOfVertices++;
			}
		}
	}
	
	fclose( fp );
	
	//sceKernelDcacheWritebackAll();
	
	mdl->displayList = (unsigned int*) memalign(16, mdl->vertTotal * sizeof(unsigned int));
	
	sceKernelDcacheWritebackAll();
	
	sceGuStart(GU_CALL, mdl->displayList);
	
	sceGuDrawArray(GU_TRIANGLES, GU_TEXTURE_32BITF|GU_VERTEX_32BITF|GU_TRANSFORM_3D, mdl->vertTotal, 0, mdl->verts);
	
	sceGuFinish();
	sceGuSync(0,0);
	
	return mdl;
}

void unloadModel(Model *mdl)
{
	free(mdl->displayList);
	free(mdl->verts);
	free(mdl);
}

void drawModel(Model *mdl, Image *img)
{
	sceGuTexImage(0,img->realWidth,img->realHeight,img->realWidth,(void*)img->data);
	sceGumLoadIdentity();
	sceGumUpdateMatrix(); //  update the matrix
	sceGuCallList(mdl->displayList);
	//sceGumDrawArray(GU_TRIANGLES, GU_TEXTURE_32BITF|GU_VERTEX_32BITF|GU_TRANSFORM_3D, mdl->vertTotal, 0, mdl->verts);
}

void drawModelPos(float x, float y, float z, Model *mdl, Image *img)
{
	sceGuTexImage(0,img->realWidth,img->realHeight,img->realWidth,(void*)img->data);
	
	sceGumLoadIdentity();
	ScePspFVector3 move = {x, y, z};
	sceGumTranslate(&move);
	
	sceGumUpdateMatrix(); //  update the matrix
	sceGuCallList(mdl->displayList);
	//sceGumDrawArray(GU_TRIANGLES, GU_TEXTURE_32BITF|GU_VERTEX_32BITF|GU_TRANSFORM_3D, mdl->vertTotal, 0, mdl->verts);
}

void setFog(float near, float far, unsigned int color)
{
	sceGuStart(GU_DIRECT, DList);
	sceGuEnable(GU_FOG);
	sceGuFog(near, far, color);
	sceGuFinish();
	sceGuSync(0,0);
}

//---------------------------------------------------------------------------------//
// Camera                                                                          //
//---------------------------------------------------------------------------------//

void setCamera3rdPerson(float posX, float posY, float posZ, float rotX, float rotY, float zoom)
{
	sceGumMatrixMode(GU_VIEW);
	sceGumLoadIdentity();
	{
		ScePspFVector3 eye = { posX - ( zoom * sin(rotY) * cos(rotX) ),
							   posY + (zoom * sin(rotX) ),
							   posZ + ( zoom * cos(rotY) * cos(rotX) ) };
		ScePspFVector3 center = {posX, posY, posZ};
		ScePspFVector3 up = {0.0f, 1.0f, 0.0f};
		
		sceGumLookAt(&eye, &center, &up);
	}
	sceGumMatrixMode(GU_MODEL);
	sceGumLoadIdentity();
}

//---------------------------------------------------------------------------------//
// Audio                                                                           //
//---------------------------------------------------------------------------------//

//---------------------------------------------------------------------------------//
// Font                                                                            //
//---------------------------------------------------------------------------------//

//---------------------------------------------------------------------------------//
// Controller                                                                      //
//---------------------------------------------------------------------------------//

CONTROLLER *button;

CONTROLLER *readInput(int masked,int init, int interval)
{
	// Get the buttons currently being pressed
	sceCtrlPeekBufferPositive(&ctrl, 1);

	button->repeatMask = masked; // Assign the buttons masks
	button->repeatInit = init; // assign the repeat init
	button->repeatInterval = interval; // assign the repeat intervals

	// if repeat counter is greater then or equal to the repeat interval and th repeat counter is equal to the button init
	if(button->repeatCounter >= button->repeatInterval && button->repeatCounter == button->repeatInit) {
		button->repeatMask = MASK_NONE; // Remove masked buttons for the rest of the frame
		button->repeatCounter -= button->repeatInterval; // subtract repeat interval from repeat counter so we get that fun little sped up effect(like with menu's)
		// if the masked buttons are the same as the held buttons, then increament the counter
	} else if(button->repeatMask & ctrl.Buttons) {
		if(button->repeatCounter == 0) button->repeatMask = MASK_NONE;
		button->repeatCounter++;
		// otherwise, set the counter to 0
	} else
		button->repeatCounter = 0;

	button->released.value = button->held.value & ~ctrl.Buttons; // assign the released buttons
	button->pressed.value = ~button->held.value & ctrl.Buttons; // assign the pressed buttons
	button->held.value = ctrl.Buttons & ~button->repeatMask; // assign the held buttons

	button->analogX = (short)ctrl.Lx-128; // assign the analog x value
	button->analogY = (short)ctrl.Ly-128; // assign the anolog y value

	return button; // return all the button info
}

//---------------------------------------------------------------------------------//
// Debugging                                                                       //
//---------------------------------------------------------------------------------//

void Stats()
{
	// FPS
	fps++;
	sceRtcGetCurrentTick( &fpsTickNow );

	if( ((fpsTickNow - fpsTickLast)/((float)tickResolution)) >= 1.0f )
	{
		fpsTickLast = fpsTickNow;
		lastFps = fps;
		fps = 0;
	}

	pspDebugScreenSetOffset((int)fbp0);
	pspDebugScreenSetXY(0,0);
	pspDebugScreenPrintf("Mem: %i	Graphics Mem: %d	FPS: %d (%dMB/s)", sceKernelMaxFreeMemSize(), 262144-sceGuCheckList(), (int)lastFps,(((int)lastFps * SCR_WIDTH * SCR_HEIGHT * 2)/(1024*1024)));
}

void StatsExtra(const char *format, ...)
{
	va_list opt = 0;
	char buff[2048];
	va_start(opt, format);
	vsnprintf( buff, (size_t) sizeof(buff), format, opt);
	
	// FPS
	fps++;
	sceRtcGetCurrentTick( &fpsTickNow );

	if( ((fpsTickNow - fpsTickLast)/((float)tickResolution)) >= 1.0f )
	{
		fpsTickLast = fpsTickNow;
		lastFps = fps;
		fps = 0;
	}

	pspDebugScreenSetOffset((int)fbp0);
	pspDebugScreenSetXY(0,0);
	pspDebugScreenPrintf("Mem: %i	Graphics Mem: %d	FPS: %d (%dMB/s)\n%s", sceKernelMaxFreeMemSize(), 262144-sceGuCheckList(), (int)lastFps,(((int)lastFps * SCR_WIDTH * SCR_HEIGHT * 2)/(1024*1024)), buff);
}
