/*

Velocity
Version 1.1
Written by Gabriel Anderson (Zettablade)
Copyright 2006, 2007 Gabriel Anderson

	This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

!!!DISCLAIMERS!!!
!!!DO NOT CLAIM VELOCITY AS YOUR OWN!!! 
!!!DO NOT USE VELOCITY TO MAKE MALICIOUS CONTENT!!!
!!!DO NOT USE VELOCITY FOR ANY COMMERCIAL PURPOSES WITHOUT EXPRESS PERMISSION FROM THE AUTHOR!!!
!!!DISCLAMERS!!!

*/

//---------------------------------------------------------------------------------//
// Includes                                                                        //
//---------------------------------------------------------------------------------//

#include <pspkernel.h>
#include <pspdisplay.h>
#include <pspdebug.h>
#include <stdlib.h>
#include <stdio.h>
#include <stdarg.h>
#include <math.h>
#include <ctype.h>
#include <string.h>
#include <pspctrl.h>
#include <pspgu.h>
#include <pspgum.h>
#include <psprtc.h>
#include <psppower.h>
#include <malloc.h>
#include <png.h>
#include <pspaudio.h>
#include <pspatrac3.h>
#include <pspvfpu.h>

#if defined(__cplusplus)
extern "C" {
#endif

//---------------------------------------------------------------------------------//
// Data                                                                            //
//---------------------------------------------------------------------------------//

extern SceCtrlData ctrl;

// Structs for Verts!

typedef struct {
	u32 color;
	short x, y, z;
} CP_Vertex_2D; // Color, Position, 2D

typedef struct {
	short u, v;
	short x, y, z;
} TP_Vertex_2D; // Texture, Position, 2D

typedef struct {
	u32 color;
	float x, y, z;
} CP_Vertex_3D; // Color, Position, 3D

typedef struct {
	float nx, ny, nz;
	u32 color;
	float x, y, z;
} NCP_Vertex_3D; // Normals, Color, Position, 3D

typedef struct {
	float u, v;
	float x, y, z;
} TP_Vertex_3D; // Texture, Position, 3D

typedef struct {
	float nx, ny, nz;
	float u, v;
	float x, y, z;
} NTP_Vertex_3D; // Normals, Texture, Position, 3D

// Holds an image
typedef struct {
	int realWidth, realHeight;	// Width and height of the texture (both pow2)
	float pixelWidth, pixelHeight; // The initial width and height of the texture in pixels
	int totalSize; // The total size of the image (bytes)
	u32* data; // image data
	unsigned int __attribute__((aligned(16))) *displayList;
} Image;

// Holds a model
typedef struct {
	int vertTotal;
	unsigned int __attribute__((aligned(16))) *displayList;
	TP_Vertex_3D *verts;
} Model;

typedef struct {
	union {
		struct {
			int select:1, start:3;
			int up:1, right:1, down:1, left:1;
			int L:1, R:1;
			int triangle:3, circle:1, cross:1, square:1;
			int home:1, hold:1, wlanOn:1, remote:1, volUp:1;
			int volDown:1, screen:1, note:1, disc:1, ms:1;
		};
		int value; // what buttons are on
	} pressed; // For checking if a button is pressed

	union {
		struct {
			int select:1, start:3;
			int up:1, right:1, down:1, left:1;
			int L:1, R:1;
			int triangle:3, circle:1, cross:1, square:1;
			int home:1, hold:1, wlanOn:1, remote:1, volUp:1;
			int volDown:1, screen:1, note:1, disc:1, ms:1;
		};
		int value; // what buttons are on
	} held; // For checking if a button is held

	union {
		struct {
			int select:1, start:3;
			int up:1, right:1, down:1, left:1;
			int L:1, R:1;
			int triangle:3, circle:1, cross:1, square:1;
			int home:1, hold:1, wlanOn:1, remote:1, volUp:1;
			int volDown:1, screen:1, note:1, disc:1, ms:1;
		};
		int value; // what buttons are on
	} released; // For checking if a button is released

	int repeatMask; // what buttons are going to be repeated
	int repeatCounter; // A counter for the button repeats when held
	int repeatInit; // The amount of time between the first repeat when held (this only happens once, unless the button is released and held again(physically))
	int repeatInterval; // The amount of time between repeats after the first repeat (this is continous, and is used until the button is released(physically))

	short analogX; // The current of the analog nub horizontally. Can be anywhere from 128 to -128.
	short analogY; // The current of the analog nub horizontally. Can be anywhere from 128 to -128.
} CONTROLLER;

extern CONTROLLER *button;

extern void* fbp0;
extern void* fbp1;
extern void* zbp;
extern unsigned int __attribute__((aligned(16))) DList[262144];

extern int fps, lastFps;		// for calculating the frames per second

extern u32 tickResolution;
extern u64 fpsTickNow;
extern u64 fpsTickLast;

extern CONTROLLER buttonInit;
extern SceCtrlData ctrl;

extern const int PRINT_BUFF;

extern const int IN_RAM;
extern const int IN_VRAM;

extern const int OFF;
extern const int ON;

extern const int DRAW_2D;
extern const int DRAW_3D;

extern const int PI;

#define RGB(r,g,b)		((r) | ((g)<<8) | ((b)<<16) | (0xff<<24))
#define RGBA(r,g,b,a)	((r) | ((g)<<8) | ((b)<<16) | ((a)<<24))

#define IsSet(val,flag) (val&flag)==flag

#define MASK_NONE 0x0
#define MASK_SELECT 0x000001
#define MASK_START 0x000008 
#define MASK_UP 0x000010
#define MASK_RIGHT 0x000020
#define MASK_DOWN 0x000040
#define MASK_LEFT 0x000080
#define MASK_LTrig 0x000100
#define MASK_RTrig 0x000200
#define MASK_TRIANGLE 0x001000
#define MASK_CIRCLE 0x002000
#define MASK_CROSS 0x004000
#define MASK_SQUARE 0x008000
#define MASK_HOME 0x010000
#define MASK_HOLD 0x020000
#define MASK_NOTE 0x800000
#define MASK_SCREEN 0x400000
#define MASK_VOLUP 0x100000
#define MASK_VOLDOWN 0x200000
#define MASK_WLANON 0x040000
#define MASK_REMOTE 0x080000
#define MASK_DISC 0x1000000
#define MASK_MS 0x2000000
#define MASK_ALL MASK_SELECT|MASK_START|MASK_UP|MASK_RIGHT|MASK_DOWN|MASK_LEFT|MASK_LTrig|MASK_RTrig|MASK_TRIANGLE|MASK_CIRCLE|MASK_CROSS|MASK_SQUARE

//---------------------------------------------------------------------------------//
// General                                                                         //
//---------------------------------------------------------------------------------//

void initVLib(u32 clearColor);	// Call this BEFORE your MAIN LOOP!!! CALL ONLY ONCE!!!
void uninitVLib();				// Call this AFTER your main loop!!! CALL ONLY ONCE!!!
void startDrawing();			// Begins the drawing routine
void endDrawing();				// Ends the drawing routine
void syncFrame();				// flips the screen and waits the vsync
void flipScreen();				// Switches the screen buffers
void clearScreen();				// Clears the screen. Downside to this is that it can hurt your fps.
void waitVSync();				// Keeps you from going over 60fps
// Switch between DRAW_2D and DRAW_3D draw modes
void setDrawMode(short drawMode);

//---------------------------------------------------------------------------------//
// Image                                                                           //
//---------------------------------------------------------------------------------//

// Loaders
Image *loadImage(const char *filename, int location); // Determines the type of image and loads it
void unloadImage(Image *img);

// Fast drawing functions (not very flexible, but much faster then regular drawing
// This pre draws the image
void preDraw(Image *img);

// Drawing Functions
void drawImage(float x, float y, float width, float height, Image *img);
void drawImageCenter(float x, float y, float width, float height, float centerX, float centerY, Image *img);
void drawImageAngle(float x, float y, float width, float height, float angle, Image *img);
void drawImageAngleCenter(float x, float y, float width, float height, float centerX, float centerY, float angle, Image *img);
void drawImageCrop(float x, float y, float width, float height, float beginX, float beginY, float endX, float endY, Image *img);
void drawImageCropCenter(float x, float y, float width, float height, float centerX, float centerY, float beginX, float beginY, float endX, float endY, Image *img);
void drawImageCropAngle(float x, float y, float width, float height, float beginX, float beginY, float endX, float endY, float angle, Image *img);
void drawImageCropAngleCenter(float x, float y, float width, float height, float centerX, float centerY, float beginX, float beginY, float endX, float endY, float angle, Image *img);
		
// Image Effects
// Flips to the image either left or right (depending on what it was before) (UNFINISHED)
void flip(Image *img);
// Allows you to set the images alpha (SLOW! Done pixel by pixel) (UNFINISHED)
void setImageAlpha(Image *img, u8 alpha);
// Smooths ALL textures. Turn off it your fps is hurting.
void linearFilter(short state);

//---------------------------------------------------------------------------------//
// 3D Model                                                                        //
//---------------------------------------------------------------------------------//

Model *loadModel(const char *filename);
void unloadModel(Model *mdl);

void drawModel(Model *mdl, Image *img);
void drawModelPos(float x, float y, float z, Model *mdl, Image *img);
void drawModelRot(float x, float y, float z, Model *mdl, Image *img); // unfinished
void drawModelPosRot(float posX, float posY, float posZ, float rotX, float rotY, float rotZ, Model *mdl, Image *img); // unfinished

void setFog(float near, float far, unsigned int color); // Need to add in a toggle to turn this on/off

//---------------------------------------------------------------------------------//
// Camera                                                                          //
//---------------------------------------------------------------------------------//

void setCamera3rdPerson(float posX, float posY, float posZ, float rotX, float rotY, float zoom);

//---------------------------------------------------------------------------------//
// Audio                                                                           //
//---------------------------------------------------------------------------------//

//---------------------------------------------------------------------------------//
// Video                                                                           //
//---------------------------------------------------------------------------------//

//---------------------------------------------------------------------------------//
// Font                                                                            //
//---------------------------------------------------------------------------------//

//---------------------------------------------------------------------------------//
// Controller                                                                      //
//---------------------------------------------------------------------------------//

CONTROLLER *readInput(int masked,int init, int interval);

//---------------------------------------------------------------------------------//
// Debugging                                                                       //
//---------------------------------------------------------------------------------//

void Stats(); // displays the current amount of free memory in bytes and the current fps.
void StatsExtra(const char *format, ...);

#if defined(__cplusplus)
}
#endif
